# --------------------------------------------------------------------------------------------------
# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2024 Jayesh Badwaik <j.badwaik@fz-juelich.de>
# --------------------------------------------------------------------------------------------------

import itertools
import os
import tempfile
import shutil
import subprocess
import json
import sys
import matplotlib.pyplot
import math
import numpy


def prepare_plotting_data(experiment_array):
    complete_plotting_data = {}
    complete_plotting_data["experiment"] = []

    for experiment in experiment_array:
        plotting_data = {}
        plotting_data["pipeline"] = experiment.pipeline()
        plotting_data["workload_factor"] = float(experiment.workload_factor())
        plotting_data["runtime"] = {}
        plotting_data["runtime"]["nodes"] = []
        plotting_data["runtime"]["runtime"] = []
        for data in experiment.json_data()["data"]:
            plotting_data["runtime"]["nodes"].append(int(data["parameter"]["nodes"]))
            plotting_data["runtime"]["runtime"].append(float(data["runtime"]))

            plotting_data["runtime"]["runtime"] = [
                x
                for _, x in sorted(
                    zip(plotting_data["runtime"]["nodes"], plotting_data["runtime"]["runtime"])
                )
            ]

            plotting_data["runtime"]["nodes"] = sorted(plotting_data["runtime"]["nodes"])

        plotting_data["label"] = experiment.prefix()
        plotting_data["system"] = experiment.json_data()["experiment"]["system"]

        complete_plotting_data["experiment"].append(plotting_data)

    min_nodes = sys.maxsize
    max_nodes = 0
    min_runtime = sys.float_info.max
    max_runtime = 0

    max_workload_factor = sys.float_info.max

    for pipeline_data in complete_plotting_data["experiment"]:
        local_max = max(pipeline_data["runtime"]["nodes"])
        local_min = min(pipeline_data["runtime"]["nodes"])
        if pipeline_data["system"] == "jedi":
            factor = 1
        else:
            factor = 0.5
        min_nodes =   min(min_nodes, local_min)
        max_nodes =   factor * max(max_nodes, local_max)
        local_max =   max(pipeline_data["runtime"]["runtime"])
        local_min =   min(pipeline_data["runtime"]["runtime"])
        min_runtime = min(min_runtime, local_min)
        max_runtime = max(max_runtime, local_max)

    complete_plotting_data["node_range"] = [0.5 * min_nodes, 1.4 * max_nodes]
    complete_plotting_data["runtime_range"] = [0.5 * min_runtime, 1.4 * max_runtime]

    workload_factor_array = [
        data["workload_factor"] for data in complete_plotting_data["experiment"]
    ]
    min_workload_factor = min(workload_factor_array)
    normalized_wf = [p / min_workload_factor for p in workload_factor_array]

    for index, pipeline_data in enumerate(complete_plotting_data["experiment"]):
        pipeline_data["expected_runtime"] = {}
        pipeline_data["expected_runtime"]["nodes"] = pipeline_data["runtime"]["nodes"]
        pipeline_data["expected_runtime"]["runtime"] = [
            normalized_wf[index] * p for p in pipeline_data["runtime"]["runtime"]
        ]

    return complete_plotting_data


def generate_plot_pdf_file(plotting_data, output_dir):
    plot = matplotlib.pyplot.figure()
    marker = itertools.cycle(("o"))

    ax = plot.subplots()
    ax.set_xlabel("Number of Nodes ($N_\\text{JEDI}$, $N_{4 \\mathrm{x} \\text{A100}}/2$; log-scale)")
    ax.set_ylabel("Runtime / s (log-scale)")
    ax.set_yscale("log")
    ax.set_xscale("log")
    ax.margins(2)
    ax.set_xlim(plotting_data["node_range"])
    ax.set_ylim(plotting_data["runtime_range"])

    xticklabels = []
    yticklabels = []

    cmap = matplotlib.colormaps["rainbow"]

    colorgen = iter(cmap(numpy.linspace(0, 1, len(plotting_data["experiment"]))))
    ax.grid(visible=True, which="major", axis="x", linestyle="-", linewidth=0.5, alpha=0.5)


    oldcolor = None
    for pipeline_data in plotting_data["experiment"]:
        color = next(colorgen)
        xticklabels = xticklabels + pipeline_data["runtime"]["nodes"]
        yticklabels = yticklabels + pipeline_data["runtime"]["runtime"]
        if pipeline_data["system"] == "jedi":
            oldcolor=color
            xaxis_data = pipeline_data["runtime"]["nodes"]
            yaxis_data = [p for p in pipeline_data["runtime"]["runtime"] ]
        else:
            yaxis_data = [p for p in pipeline_data["runtime"]["runtime"] ]
            xaxis_data = [p / 2 for p in pipeline_data["runtime"]["nodes"]]


        ideal_scaling_array = []
        low_scaling_array = []
        for i in range(len(xaxis_data)):
            ideal_scaling = yaxis_data[0] / xaxis_data[i] * xaxis_data[0]
            low_scaling = ideal_scaling * 1.25
            ideal_scaling_array.append(ideal_scaling)
            low_scaling_array.append(low_scaling)

        system = pipeline_data["system"]
        prefix = pipeline_data["pipeline"]
        ax.plot(xaxis_data, yaxis_data, label=system, marker=next(marker), color=color, markersize=4)

        if pipeline_data["system"] != "jedi":
            error_color = color
        else:
            error_color = color

        ax.plot(xaxis_data, ideal_scaling_array, color=error_color, ls="--", lw=0.6, alpha=1)
        ax.plot(xaxis_data, low_scaling_array, color=error_color, ls="--", lw=0.6, alpha=1)


        # if float(pipeline_data["workload_factor"]) > 1:
        #    expected_yaxis = pipeline_data["expected_runtime"]["runtime"]
        #    error_yaxis = [p * 0.2 for p in expected_yaxis]
        #    ax.errorbar(xaxis_data, yaxis_data, yerr=error_yaxis, fmt="o")

    xticklabels = sorted(list(set(xticklabels)))
    yticklabels = sorted(list(set(yticklabels)))

    min_yticklabel = math.log(min(yticklabels))
    max_yticklabel = math.log(max(yticklabels))

    yticklabels = numpy.logspace(min_yticklabel, max_yticklabel, num=6, base=math.e)
    yticklabels = [int(p) for p in yticklabels]

    ax.set_xticks(xticklabels)
    ax.set_xticklabels(xticklabels)
    ax.set_yticks(yticklabels)
    ax.set_yticklabels(yticklabels)

    plot.legend(bbox_to_anchor=(0.5, 0.3))

    plot.savefig(
        os.path.join(
            output_dir,
            "plot.pdf",
        ),
        bbox_inches="tight",
    )
    plot.savefig(
        os.path.join(
            output_dir,
            "plot.png",
        ),
        bbox_inches="tight",
    )
    return "plot.pdf"


def generate_plot_tex_file(experiment_array, output_dir):
    plotting_data = prepare_plotting_data(experiment_array)
    pdf_filename = generate_plot_pdf_file(plotting_data, output_dir)

    plotfilename = os.path.join(output_dir, "plot.tex")

    with open(plotfilename, "w") as plotfile:
        plotfile.write("% This file was generated by jureap.\n")
        plotfile.write("\\exacbplot{" + pdf_filename + "}{Caption}\n")


def generate_csv_table_tex_file(experiment_array, output_dir):
    tablefilename = os.path.join(output_dir, "table.tex")
    with open(tablefilename, "w") as tablefile:
        tablefile.write("% This file was generated by jureap.\n")

        for experiment in experiment_array:
            csv_file = os.path.join(
                "data", experiment.output_pipeline_dir(), experiment.prefix() + ".csv"
            )
            tablefile.write("\\exacbtable{" + csv_file + "}{Caption}\n")


def generate_json_tex_file(experiment_array, output_dir):
    jsonfilename = os.path.join(output_dir, "json.tex")
    with open(jsonfilename, "w") as jsonfile:
        jsonfile.write("% This file was generated by jureap.\n")

        for experiment in experiment_array:
            json_file = os.path.join(
                "data", experiment.output_pipeline_dir(), experiment.prefix() + ".json"
            )
            jsonfile.write("\\lstinputlisting[caption=Caption]{" + json_file + "}\n")


def generate_author_tex_file(output_dir):
    authorfilename = os.path.join(output_dir, "author.tex")
    with open(authorfilename, "w") as authorfile:
        authorfile.write("% This file was generated by jureap.\n")
        authorfile.write("\\title{JEDI Evaluation Report}\n")


def compile_report_pdf(output_dir):
    subprocess.run(["make", "debug"], cwd=output_dir, env=os.environ)


def prepare_report_dir(output_dir, share_dir):
    texdir = os.path.join(share_dir, "jureap/tex/jedi")
    shutil.copytree(texdir, output_dir)


def write_json_data(experiment_array, output_dir):
    json_dir = os.path.join(output_dir, "json")
    os.makedirs(json_dir, exist_ok=True)
    for experiment in experiment_array:
        json_filepath = os.path.join(
            json_dir, experiment.pipeline() + "." + experiment.prefix() + ".json"
        )
        with open(json_filepath, "w") as jsonfile:
            json.dump(experiment.json_repr(), jsonfile, indent=4)


def copy_raw_data(input_dir, experiment_array, output_dir):
    data_dir = os.path.join(output_dir, "data")
    os.makedirs(data_dir, exist_ok=True)
    for experiment in experiment_array:
        output_experiment_reldir = experiment.pipeline() + str(".") + experiment.prefix()
        output_experiment_dir = os.path.join(data_dir, output_experiment_reldir)
        input_experiment_dir = os.path.join(input_dir, experiment.pipeline_dir())
        csv_filepath = os.path.join(input_experiment_dir, experiment.prefix() + ".csv")
        json_filepath = os.path.join(input_experiment_dir, experiment.prefix() + ".json")
        os.makedirs(output_experiment_dir, exist_ok=True)
        shutil.copy(csv_filepath, output_experiment_dir)
        shutil.copy(json_filepath, output_experiment_dir)


def generate(input_dir, experiment_array, output_dir, share_dir, tmp_dir):
    prepare_report_dir(output_dir, share_dir)
    copy_raw_data(input_dir, experiment_array, output_dir)
    generate_plot_tex_file(experiment_array, output_dir)
    generate_csv_table_tex_file(experiment_array, output_dir)
    generate_json_tex_file(experiment_array, output_dir)
    write_json_data(experiment_array, output_dir)
    generate_author_tex_file(output_dir)
    compile_report_pdf(output_dir)



# --------------------------------------------------------------------------------------------------
# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2024 Jayesh Badwaik <j.badwaik@fz-juelich.de>
# --------------------------------------------------------------------------------------------------

import sys
import os
import argparse
import jureap
import ast


def make_wide(formatter, w=140, h=100):
    """Return a wider HelpFormatter, if possible."""
    try:
        # https://stackoverflow.com/a/5464440
        # beware: "Only the name of this class is considered a public API."
        kwargs = {"width": w, "max_help_position": h}
        formatter(None, **kwargs)
        return lambda prog: formatter(prog, **kwargs)
    except TypeError:
        warnings.warn("argparse help formatter failed, falling back.")
        return formatter


def log_parser(sp):
    sp.add_argument(
        "--input-version",
        type=jureap.metadata.major_version_type,
        default=jureap.metadata.default_input_version,
        choices=jureap.metadata.supported_input_version_array,
        help="Version of Input File",
    )
    sp.add_argument(
        "--output-version",
        type=jureap.metadata.major_version_type,
        default=jureap.metadata.default_output_version,
        choices=jureap.metadata.supported_output_version_array,
        help="Version of Log File",
    )
    sp.add_argument("--input", type=str, help="Input File")
    sp.add_argument("--output", type=str, help="Output File")
    sp.add_argument("--env", type=str, default=None, help="Environment File")


def report_parser(rp):
    rp.add_argument("--input", required=True, type=str, help="Input Directory")
    rp.add_argument("--pipeline", required=True, type=str, help="Pipeline Array")
    rp.add_argument("--prefix", required=True, type=str, help="Prefix Array")
    rp.add_argument("--workload", required=True, type=str, help="Workload Factor Array")
    rp.add_argument("--output", required=True, type=str, help="Output Report Directory")
    rp.add_argument("--tmp", type=str, help="Temporary Directory", default=os.getcwd())
    rp.add_argument("--type", choices=["jedi", "scaling"], required=True, help="Report Type")


def check_parser(sp):
    sp.add_argument("input_file", type=str, help="Result File")


def top_level():
    parser = argparse.ArgumentParser(
        description="jureap",
        formatter_class=make_wide(argparse.ArgumentDefaultsHelpFormatter),
    )
    subparsers = parser.add_subparsers(dest="subcommand", required=True)
    lp = subparsers.add_parser(
        "log",
        help="Log Generator",
        formatter_class=make_wide(argparse.ArgumentDefaultsHelpFormatter),
    )
    rp = subparsers.add_parser("report", help="Report Generator")

    vp = subparsers.add_parser("check", help="check Result File")

    log_parser(lp)
    check_parser(vp)
    report_parser(rp)
    return parser


def main():
    bin_dir = os.path.dirname(os.path.realpath(sys.argv[0]))
    root_dir = os.path.dirname(bin_dir)
    share_dir = os.path.join(root_dir, "share")

    try:
        raw_args = sys.argv[1:]
        parser = top_level()
        args = parser.parse_args(raw_args)

        if args.subcommand == "check":
            input_filename = args.input_file
            jureap.check.check(input_filename)
        elif args.subcommand == "report":
            pipeline_array = ast.literal_eval(args.pipeline)
            prefix_array = ast.literal_eval(args.prefix)
            workload_factor_array = ast.literal_eval(args.workload)

            jureap.report.generate(
                args.type,
                args.input,
                pipeline_array,
                prefix_array,
                workload_factor_array,
                args.output,
                share_dir,
                args.tmp,
            )
        elif args.subcommand == "log":
            if args.env == None:
                environment = os.environ
            else:
                environment = jureap.detail.env_file_to_dict(args.env)

            jureap.log.log(
                args.input, args.output, environment, args.input_version, args.output_version
            )
        else:
            print("Usage Error: Subcommand " + args.subcommand + " not implemented yet.")
            exit(os.EX_USAGE)
    except Exception as e:
        print("Data Error: " + str(e))
        exit(os.EX_DATAERR)

    exit(os.EX_OK)



# --------------------------------------------------------------------------------------------------
# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2024 Jayesh Badwaik <j.badwaik@fz-juelich.de>
# --------------------------------------------------------------------------------------------------

import jureap
import pytest


def test_env_file_to_dict():
    env_file = "test/share/env/booster.env.file"
    env_dict = jureap.detail.env_file_to_dict(env_file)

    assert env_dict["CI_COMMIT_SHORT_SHA"] == """b2af1b22"""
    assert env_dict["SCRATCH_cjsc"] == """/p/scratch/cjsc"""
    assert env_dict["CI_RUNNER_EXECUTABLE_ARCH"] == """linux/amd64"""
    assert env_dict["CI_COMMIT_TITLE"] == """- fix errors"""
    assert (
        env_dict["LD_LIBRARY_PATH"]
        == """/p/software/juwelsbooster/stages/2024/software/CUDA/12/nvvm/lib64:/p/software/juwelsbooster/stages/2024/software/CUDA/12/extras/CUPTI/lib64:/p/software/juwelsbooster/stages/2024/software/CUDA/12/lib:/p/software/juwelsbooster/stages/2024/software/binutils/2.40-GCCcore-12.3.0/lib:/p/software/juwelsbooster/stages/2024/software/zlib/1.2.13-GCCcore-12.3.0/lib:/p/software/juwelsbooster/stages/2024/software/GCCcore/12.3.0/lib64"""
    )
    assert env_dict["CI_DEPENDENCY_PROXY_SERVER"] == """gitlab.jsc.fz-juelich.de:443"""
    assert env_dict["CI_PROJECT_TITLE"] == """jacamar-tester"""
    assert env_dict["CI_RUNNER_TAGS"] == """["shell", "jacamar", "login", "juwels_booster"]"""
    assert (
        env_dict["CI_PIPELINE_URL"]
        == """https://gitlab.jsc.fz-juelich.de/exacb/examples/jacamar-tester/-/pipelines/221082"""
    )



# --------------------------------------------------------------------------------------------------
# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2024 Jayesh Badwaik <j.badwaik@fz-juelich.de>
# --------------------------------------------------------------------------------------------------
import semver as SemVer
import enum


class major_version_type(enum.Enum):
    v0 = "0"
    v1 = "1"

    def __str__(self):
        return self.value


__version__ = "0.4.8"
semver = SemVer.VersionInfo.parse(__version__)
supported_input_version_array = [1]
supported_output_version_array = [1]
default_input_version = 1
default_output_version = 1



# --------------------------------------------------------------------------------------------------
# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2024 Jayesh Badwaik <j.badwaik@fz-juelich.de>
# --------------------------------------------------------------------------------------------------

import jureap.detail
import csv


def env_file_to_dict(env_file):
    """
    Convert the environment file to a dictionary.
    :param env_file: Path to the environment file.
    :return: Dictionary containing the environment variables.
    """
    env_dict = {}
    with open(env_file, "r") as f:
        for line in f:
            if line.strip() == "":
                continue
            key, value = line.strip().split("=", 1)
            env_dict[key] = value
    return env_dict


def csv_file_to_array(input_filename):
    matrix = []

    with open(input_filename, "r") as input_file:
        reader = csv.reader(input_file)
        for row in reader:
            matrix.append(row)

    output = []
    column_array = matrix[0]

    for row in matrix[1:]:
        if len(row) == 0:
            continue
        row_dict = {}
        for index, value in enumerate(row):
            row_dict[column_array[index]] = value
        output.append(row_dict)
    return output


def parse_reporter_info(environment):
    reporter_info = {}
    reporter_info["pipeline"] = environment["CI_PIPELINE_ID"]
    reporter_info["job"] = environment["CI_JOB_ID"]
    reporter_info["commit"] = environment["CI_COMMIT_SHA"]
    reporter_info["username"] = environment["GITLAB_USER_LOGIN"]
    reporter_info["project"] = environment["PROJECT"]
    reporter_info["budget"] = environment["BUDGET_ACCOUNTS"]
    reporter_info["system"] = environment["SYSTEMNAME"]
    reporter_info["version"] = "2024.12"
    reporter_info["timestamp"] = environment["CI_JOB_STARTED_AT"]
    return reporter_info



# --------------------------------------------------------------------------------------------------
# SPDX-License-Identifier: Apache-2.0
# Copyright (C) 2024 Jayesh Badwaik <j.badwaik@fz-juelich.de>
# --------------------------------------------------------------------------------------------------

import jureap.check
import os
import argparse
import csv
import json
import jureap.metadata
import pytest


def test_00_check_csv():
    input_filename = "test/share/check/00.input.csv"
    jureap.check.check(input_filename)


def test_01_check_csv_fail():
    input_filename = "test/share/check/01.input.fail.csv"

    with pytest.raises(ValueError, match="Empty value for column"):
        jureap.check.check(input_filename)


def test_02_check_csv_fail():
    input_filename = "test/share/check/02.input.fail.csv"

    with pytest.raises(ValueError, match="Column .* not found in the input file"):
        jureap.check.check(input_filename)
